# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import kubernetes
from .pvc import get_storageclass
from .proxy import service_proxy


TENSORBOARD_PREFIX = "adaptdl-tensorboard-"


def tensorboard_create(args, remaining):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    name = "{}{}".format(TENSORBOARD_PREFIX, args.name)
    labels = {"adaptdl/tensorboard": args.name}
    storageclass = get_storageclass(name=args.storageclass)
    # Create Deployment.
    container = kubernetes.client.V1Container(
        name="tensorboard",
        image="tensorflow/tensorflow",
        command=["tensorboard"],
        args=["--host=0.0.0.0", "--logdir=/adaptdl/tensorboard"],
        ports=[kubernetes.client.V1ContainerPort(container_port=6006)],
        volume_mounts=[
            kubernetes.client.V1VolumeMount(
                name="logdir",
                mount_path="/adaptdl/tensorboard",
            ),
        ],
    )
    template = kubernetes.client.V1PodTemplateSpec(
        metadata=kubernetes.client.V1ObjectMeta(labels=labels),
        spec=kubernetes.client.V1PodSpec(
            volumes=[
                kubernetes.client.V1Volume(
                    name="logdir",
                    persistent_volume_claim=kubernetes.
                    client.V1PersistentVolumeClaimVolumeSource(
                       claim_name=name),
                )
            ],
            containers=[container],
        )
    )
    deployment = kubernetes.client.V1Deployment(
        metadata=kubernetes.client.V1ObjectMeta(
            namespace=namespace,
            name=name,
            labels=labels,
        ),
        spec=kubernetes.client.V1DeploymentSpec(
            selector=kubernetes.client.V1LabelSelector(match_labels=labels),
            template=template,
        ),
    )
    apps_api = kubernetes.client.AppsV1Api()
    deployment = apps_api.create_namespaced_deployment(namespace, deployment)
    # Construct OwnerReference.
    owner_reference = kubernetes.client.V1OwnerReference(
        api_version="v1",
        kind="Deployment",
        name=deployment.metadata.name,
        uid=deployment.metadata.uid,
    )
    # Create PersistentVolumeClaim.
    claim = kubernetes.client.V1PersistentVolumeClaim(
        metadata=kubernetes.client.V1ObjectMeta(
            name=name,
            labels=labels,
            owner_references=[owner_reference],
        ),
        spec=kubernetes.client.V1PersistentVolumeClaimSpec(
            storage_class_name=storageclass.metadata.name,
            access_modes=["ReadWriteMany"],
            volume_mode="Filesystem",
        ),
    )
    if args.size is not None:
        claim.spec.resources = kubernetes.client.V1ResourceRequirements(
            requests={"storage": args.size})
    core_api = kubernetes.client.CoreV1Api()
    core_api.create_namespaced_persistent_volume_claim(namespace, claim)
    # Create Service.
    service = kubernetes.client.V1Service(
        metadata=kubernetes.client.V1ObjectMeta(
            name=name,
            labels=labels,
            owner_references=[owner_reference],
        ),
        spec=kubernetes.client.V1ServiceSpec(
            selector=labels,
            type=("NodePort" if args.nodeport else "ClusterIP"),
            ports=[
                kubernetes.client.V1ServicePort(
                    port=6006,
                    target_port=6006,
                ),
            ],
        )
    )
    core_api.create_namespaced_service(namespace, service)
    print("Successfully created TensorBoard instance {}".format(args.name))


def tensorboard_delete(args, remaining):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    name = "{}{}".format(TENSORBOARD_PREFIX, args.name)
    apps_api = kubernetes.client.AppsV1Api()
    apps_api.delete_namespaced_deployment(name, namespace)
    print("Successfully deleted TensorBoard instance {}".format(args.name))


def tensorboard_list(args, remaining):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    apps_api = kubernetes.client.AppsV1Api()
    core_api = kubernetes.client.CoreV1Api()
    deployment_list = apps_api.list_namespaced_deployment(
        namespace, label_selector="adaptdl/tensorboard")
    pvc_list = core_api.list_namespaced_persistent_volume_claim(
        namespace, label_selector="adaptdl/tensorboard")
    svc_list = core_api.list_namespaced_service(
        namespace, label_selector="adaptdl/tensorboard")
    records = []
    for deployment in deployment_list.items:
        pvc = next((pvc for pvc in pvc_list.items
                    if pvc.metadata.name == deployment.metadata.name), None)
        svc = next((svc for svc in svc_list.items
                    if svc.metadata.name == deployment.metadata.name), None)
        records.append({
            "NAME": deployment.metadata.labels["adaptdl/tensorboard"],
            "NODEPORT": svc and svc.spec.ports[0].node_port,
            "SIZE": pvc and pvc.spec.resources.requests.get("storage"),
            "READY": "{}/{}".format(deployment.status.ready_replicas or 0,
                                    deployment.status.replicas or 0),
        })
    # Format and print table.
    header = ["NAME", "NODEPORT", "SIZE", "READY"]
    widths = [len(key) + 3 for key in header]
    for rec in records:
        for idx, key in enumerate(header):
            rec[key] = "<none>" if rec[key] is None else str(rec[key])
            widths[idx] = max(widths[idx], len(rec[key]) + 3)
    line = "".join("{:<" + str(w) + "}" for w in widths)
    print(line.format(*header))
    for rec in records:
        print(line.format(*[rec[key] for key in header]))


def tensorboard_proxy(args, remaining):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    service = f"adaptdl-tensorboard-{args.name}"
    with service_proxy(namespace, service, args.address, args.port) as addr:
        print(f"Proxying to TensorBoard instance {args.name} at http://{addr}")
        try:
            while True:
                time.sleep(1000000)
        except KeyboardInterrupt:
            pass


def add_tensorboard_commands(parser):
    parser.set_defaults(handler=lambda args, remaining: parser.print_help())
    subparsers = parser.add_subparsers(help="sub-command help")
    # create
    parser_create = subparsers.add_parser(
        "create", help="create tensorboard deployment")
    parser_create.add_argument(
        "name", type=str, help="tensorboard deployment name")
    parser_create.add_argument("--nodeport", action="store_true",
                               help="also create NodePort service")
    parser_create.add_argument(
        "--storageclass", type=str, help="name of StorageClass")
    parser_create.add_argument(
        "--size", type=str, default="1Gi", help="storage size")
    parser_create.set_defaults(handler=tensorboard_create)
    # delete
    parser_delete = subparsers.add_parser(
        "delete", help="delete tensorboard deployment")
    parser_delete.add_argument(
        "name", type=str, help="tensorboard deployment name")
    parser_delete.set_defaults(handler=tensorboard_delete)
    # list
    parser_list = subparsers.add_parser(
        "list", help="list tensorboard deployments")
    parser_list.set_defaults(handler=tensorboard_list)
    # proxy
    parser_proxy = subparsers.add_parser(
       "proxy", help="proxy to a tensorboard deployment")
    parser_proxy.add_argument(
       "name", type=str, help="tensorboard deployment name")
    parser_proxy.add_argument("--address", type=str, default="127.0.0.1",
                              help="local address to bind for the proxy")
    parser_proxy.add_argument("-p", "--port", type=int,
                              help="local port to bind for the proxy")
    parser_proxy.set_defaults(handler=tensorboard_proxy)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import kubernetes
import re

SUPPORTED_PROVISIONERS = (r'microk8s.io/hostpath',
                          r'.*cephfs.csi.ceph.com',
                          r'\befs\b')


def get_storageclass(name=None):
    api = kubernetes.client.StorageV1Api()
    if name is not None:
        return api.read_storage_class(name)
    # Find default storageclass.
    sc_list = api.list_storage_class()
    for sc in sc_list.items:
        for provisioner in SUPPORTED_PROVISIONERS:
            if re.search(provisioner, sc.provisioner):
                return sc
    raise SystemExit("Unsupported storageclass from available storageclasses "
                     f"{[sc.metadata.name for sc in sc_list.items]}")


def create_pvc(name=None, storage_class=None, size="100Gi",
               owner_metadata=None):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    core_api = kubernetes.client.CoreV1Api()
    if storage_class is None:
        storage_class_name = get_storageclass().metadata.name
    else:
        storage_class_name = storage_class.metadata.name
    if owner_metadata is None:
        owner_references = []
    else:
        owner_references = [
            kubernetes.client.V1OwnerReference(
                api_version="adaptdl.petuum.com/v1",
                kind="AdaptDLJob",
                name=owner_metadata["name"],
                uid=owner_metadata["uid"])]
    if name is None:
        metadata = kubernetes.client.V1ObjectMeta(
            namespace=namespace,
            generate_name="adaptdl-pvc-",
            owner_references=owner_references
        )
    else:
        metadata = kubernetes.client.V1ObjectMeta(
            namespace=namespace,
            name=name,
            owner_references=owner_references
        )
    claim = kubernetes.client.V1PersistentVolumeClaim(
        metadata=metadata,
        spec=kubernetes.client.V1PersistentVolumeClaimSpec(
            storage_class_name=storage_class_name,
            access_modes=["ReadWriteMany"],
            volume_mode="Filesystem",
            resources=kubernetes.client.V1ResourceRequirements(
                requests={"storage": size})
        ),
    )
    return core_api.create_namespaced_persistent_volume_claim(
        namespace, claim)


def create_copy_pod(pvc_name, cp_job_uid):
    context = kubernetes.config.list_kube_config_contexts()[1]
    namespace = context["context"].get("namespace", "default")
    core_api = kubernetes.client.CoreV1Api()
    volume_name = "adaptdl-pvc"

    labels = {"adaptdl/cli-copy": cp_job_uid}

    pvc = core_api.read_namespaced_persistent_volume_claim(pvc_name, namespace)
    pvc_uid = pvc.metadata.uid

    metadata = {"namespace": namespace,
                "generateName": "copy-{}".format(pvc_name),
                "labels": labels,
                "owner_references": [
                    kubernetes.client.V1OwnerReference(
                        api_version="v1",
                        kind="PersistentVolumeClaim",
                        name=pvc_name,
                        uid=pvc_uid)]
                }

    container = kubernetes.client.V1Container(
        name="copy-container",
        image="alpine",
        command=["sleep"],
        args=["1000000"],
        volume_mounts=[
            kubernetes.client.V1VolumeMount(
                name=volume_name,
                mount_path="adaptdl_pvc",
            ),
        ],
    )

    body = kubernetes.client.V1Pod(
        metadata=metadata,
        spec=kubernetes.client.V1PodSpec(
            volumes=[{
                "name": volume_name,
                "persistentVolumeClaim": {
                    "claimName": pvc_name
                    }
                }],
            containers=[container]
        )
    )
    return core_api.create_namespaced_pod(namespace, body)



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import setuptools


def read_requirements(file_path):
    requirements = []
    with open(file_path) as f:
        for line in f:
            if "#" in line:
                line = line[:line.index("#")]
            line = line.strip()
            if line and not line.startswith("-"):
                requirements.append(line)
    return requirements


if __name__ == "__main__":
    setuptools.setup(
        name="adaptdl-cli",
        version=os.getenv("ADAPTDL_VERSION", "0.0.0"),
        author="Petuum Inc. & The AdaptDL Authors",
        author_email="aurick.qiao@petuum.com",
        description="Dynamic-resource trainer and scheduler for deep learning",
        url="https://github.com/petuum/adaptdl",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: Other/Proprietary License",
            "Operating System :: POSIX :: Linux",
        ],
        packages=setuptools.find_packages(include=["adaptdl_cli",
                                                   "adaptdl_cli.*"]),
        python_requires='>=3.6',
        install_requires=read_requirements("requirements.txt"),
        scripts=["bin/adaptdl"]
    )



# Copyright 2020 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from multiprocessing import Process, Event
from tempfile import NamedTemporaryFile

import kubernetes

from mitmproxy.options import Options
from mitmproxy.proxy.config import ProxyConfig
from mitmproxy.proxy.server import ProxyServer
from mitmproxy.tools.dump import DumpMaster
from portpicker import pick_unused_port
from urllib.parse import urlparse


@contextmanager
def service_proxy(namespace, service, listen_host="127.0.0.1",
                  listen_port=None, verbose=False):
    """
    This is a context manager that runs a background proxy to a Kubernetes
    service, for the duration of the managed context. The local Kubernetes
    context should have access to the /proxy verb of the given service.

    For example,

    .. code-block:: python

       with service_proxy("default", "my-service", listen_port=8080) as addr:
           print(addr)  # Should print: 127.0.0.1:8080
           # In this block, access my-service using 127.0.0.1:8080.

    Arguments:
        namespace (str): namespace of the target Kubernetes service.
        service (str): name of the target Kubernetes service, in the form
            [https:]service_name[:port_name].
        listen_host (str): local address to bind the proxy.
        listen_port (int): local port to bind the proxy. If None, selects an
            arbitrary free port.
        verbose (bool): if True, prints extra logs from the background proxy.
    """
    listen_port = pick_unused_port() if listen_port is None else listen_port
    event = Event()
    child = Process(
        target=_run_proxy, name="mitmproxy",
        args=(event, namespace, service, listen_host, listen_port, verbose),
    )
    child.start()
    event.wait()  # Wait for child to bind the port.
    try:
        yield f"{listen_host}:{listen_port}"
    finally:
        child.terminate()
        child.join()


def _run_proxy(event, namespace, service, listen_host, listen_port, verbose):
    # Run mitmproxy as a reverse-proxy to the Kubernetes ApiServer under the
    # subpath /api/v1/namespaces/{namespace}/services/{service}/proxy. See
    # https://k8s.io/docs/tasks/administer-cluster/access-cluster-services.
    client = kubernetes.config.new_client_from_config()
    prefix = f"/api/v1/namespaces/{namespace}/services/{service}/proxy"
    options = Options(listen_host=listen_host, listen_port=listen_port,
                      mode=f"reverse:{client.configuration.host}")
    master = DumpMaster(options, with_termlog=verbose, with_dumper=verbose)
    options.keep_host_header = True
    options.ssl_insecure = True
    with NamedTemporaryFile() as cert_file:
        # If Kubernetes client has a client-certificate set up, then configure
        # mitmproxy to use it. Kubernetes stores the cert and key in separate
        # files, while mitmproxy expects a single file. So append the two file
        # contents into a new file and pass it to mitmproxy.
        if client.configuration.cert_file and client.configuration.key_file:
            with open(client.configuration.cert_file, "rb") as f:
                cert_file.write(f.read())
            with open(client.configuration.key_file, "rb") as f:
                cert_file.write(f.read())
            cert_file.flush()
            options.client_certs = cert_file.name
        master.server = ProxyServer(ProxyConfig(options))
        master.addons.add(_Addon(client, prefix))
        event.set()  # Port is bound by this time, unblock parent process.
        master.run()


class _Addon(object):
    def __init__(self, client, prefix):
        self.client = client
        self.prefix = prefix

    def requestheaders(self, flow):
        # Modify outgoing request headers.
        flow.request.stream = True  # Stream for better performance.
        flow.request.path = self.prefix + flow.request.path
        self.client.update_params_for_auth(  # Inject Kubernetes authorization.
            flow.request.headers, flow.request.query, ["BearerToken"])

    def responseheaders(self, flow):
        # Modify incoming response headers.
        flow.response.stream = True  # Stream for better performance.
        if "location" in flow.response.headers:
            # Kubernetes ApiServer re-writes the Location header, undo it here.
            # See https://github.com/kubernetes/kubernetes/pull/52556.
            url = urlparse(flow.response.headers["location"])
            if url.path.startswith(self.prefix):
                url = url._replace(path=url.path[len(self.prefix):])
            flow.response.headers["location"] = url.geturl()


if __name__ == "__main__":
    import argparse
    import subprocess
    import time
    parser = argparse.ArgumentParser()
    parser.add_argument("namespace", type=str)
    parser.add_argument("service", type=str)
    parser.add_argument("--listen-host", type=str, default="127.0.0.1")
    parser.add_argument("-p", "--listen-port", type=int)
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("command", nargs=argparse.REMAINDER)
    args = parser.parse_args()
    with service_proxy(args.namespace, args.service,
                       listen_host=args.listen_host,
                       listen_port=args.listen_port,
                       verbose=args.verbose) as addr:
        if args.verbose:
            print(f"Proxy to {args.namespace}/{args.service} at {addr}")
        try:
            if args.command:
                subprocess.check_call(args.command)
            else:
                while True:
                    time.sleep(1000000)
        except (KeyboardInterrupt, subprocess.CalledProcessError):
            raise SystemExit
